from __future__ import annotations

import argparse
import logging
from collections import UserDict
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
import re
import shlex
from typing import Dict, Optional, Sequence

import argcomplete
import xdg
from dataclasses_json import DataClassJsonMixin, config
from logging_actions import log_level_action

logger = logging.getLogger(__name__)


@dataclass
class Value:
    value: str
    created: datetime = field(
        default_factory=datetime.now,
        metadata=config(encoder=datetime.isoformat, decoder=datetime.fromisoformat),
    )
    comment: Optional[str] = None


@dataclass
class Database(DataClassJsonMixin):
    mapping: Dict[str, Value]

    @classmethod
    def load_or_default(cls, file_path: Path) -> Database:
        try:
            db = cls.from_json(file_path.read_text())
        except Exception as e:
            logger.error(
                f"Unable to load from {file_path} ({e}), ignoring and using a new kv store",
            )
            db = Database(mapping={})
        return db


class LenientChoices(UserDict):
    def __contains__(self, key: object) -> bool:
        """Return True so that argparse will validate any key as a choice"""
        return True


def kv(
    kv_file: Path = xdg.xdg_data_home().joinpath("kv.json"),
    args: Optional[Sequence[str]] = None,
):
    logger.addHandler(logging.StreamHandler())
    logger.debug(f"{kv_file=}")

    db = Database.load_or_default(kv_file)

    parser = argparse.ArgumentParser(
        description="A persistent key-value store for the shell"
    )
    parser.add_argument("--log-level", action=log_level_action(logger), default="info")

    parser.add_argument("key", choices=LenientChoices(db.mapping), metavar="key")
    parser.add_argument("value", nargs="?")

    argcomplete.autocomplete(parser)
    args = parser.parse_args(args=args)

    logger.debug(f"{args=}")

    if args.value is None:
        print(db.mapping[args.key].value)
    else:
        db.mapping[args.key] = Value(value=args.value)
        kv_file.parent.mkdir(parents=True, exist_ok=True)
        kv_file.write_text(db.to_json())


def is_valid_env_var_name(s: str) -> bool:
    return re.match(pattern=r"^[a-zA-Z_]+[a-zA-Z0-9_]*$", string=s) is not None


def kvv(
    kv_file: Path = xdg.xdg_data_home().joinpath("kv.json"),
    args: Optional[Sequence[str]] = None,
):
    logger.addHandler(logging.StreamHandler())
    logger.debug(f"{kv_file=}")

    db = Database.load_or_default(kv_file)

    parser = argparse.ArgumentParser(
        description="A persistent key-value store for the shell - verbose functionality"
    )
    parser.add_argument("--log-level", action=log_level_action(logger), default="info")
    subparsers = parser.add_subparsers(required=True, dest="subcommand")

    get = subparsers.add_parser("get")
    get.add_argument("key", choices=db.mapping)

    set = subparsers.add_parser("set")
    set.add_argument("key", choices=LenientChoices(db.mapping), metavar="key")
    set.add_argument("value")
    set.add_argument("comment", nargs="?")

    remove = subparsers.add_parser("remove", aliases=["rm"])
    remove.add_argument("key", choices=db.mapping, nargs="+")

    clear = subparsers.add_parser("clear")

    env = subparsers.add_parser("env")

    argcomplete.autocomplete(parser)
    args = parser.parse_args(args=args)

    logger.debug(f"{args=}")

    if args.subcommand == "get":
        print(db.mapping[args.key].value)
    elif args.subcommand == "set":
        db.mapping[args.key] = Value(value=args.value, comment=args.comment)
        kv_file.write_text(db.to_json())
    elif args.subcommand == "remove" or args.subcommand == "rm":
        for key in args.key:
            db.mapping.pop(key)
        kv_file.write_text(db.to_json())
    elif args.subcommand == "clear":
        kv_file.write_text(Database(mapping={}).to_json())
    elif args.subcommand == "env":
        for key, value in db.mapping.items():
            if is_valid_env_var_name(key):
                print(f"{key}={shlex.quote(value.value)}")
            else:
                logger.info(f"Skipping invalid key {key}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kvsh']

package_data = \
{'': ['*']}

install_requires = \
['argcomplete>=1.12.3,<2.0.0',
 'argparse>=1.4.0,<2.0.0',
 'dataclasses-json>=0.5.6,<0.6.0',
 'jsonschema>=4.2.1,<5.0.0',
 'logging-actions>=0.1.6,<0.2.0',
 'xdg>=5.1.1,<6.0.0']

entry_points = \
{'console_scripts': ['kv = kvsh:kv', 'kvv = kvsh:kvv']}

setup_kwargs = {
    'name': 'kvsh',
    'version': '0.1.2',
    'description': 'Persistent key-value store for the command line',
    'long_description': '# kvsh, a command-line key-value store for the shell\n- Cross-shell persistence\n- Easy UI, with tab completion\n## `kv` for quick operations\n```shell\n$ kv hello world\n$ kv hello\nworld\n```\n## `kvv` for advanced functionality\n```shell\n$ kv hello world\n$ kv foo food\n$ kv bar bartender\n$ kvv env # Print key=value iff key is a valid environment variable name\nhello=world\nfoo=food\nbar=bartender\n$ kvv remove hello\n$ kvv env\nhello=world\n$ kvv clear\n```\n## Installation\nRecommended installation with [`pipx`](https://github.com/pypa/pipx):\n```shell\npipx install kvsh\n```\nTab completion with [`argcomplete`](https://github.com/kislyuk/argcomplete):\n```shell\npipx install argcomplete\neval "$(register-python-argcomplete kv)"\neval "$(register-python-argcomplete kvv)"\n```',
    'author': 'Aatif Syed',
    'author_email': 'aatifsyedyp@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aatifsyed/kvsh',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
